//
//  IDatabase.m
//  ExamplePersistSQLite
//
//  Created by Fernando Nazario on 14/07/11.
//  Copyright 2011 __MyCompanyName__. All rights reserved.
//

#import "IDatabase.h"

@implementation IDatabase

@synthesize pathToDatabase;

static IDatabase *_shared = nil;

+ (IDatabase *)shared
{
	@synchronized([IDatabase class])
	{
		if (!_shared) {
			_shared = [[self alloc] init];
		}
	}
	return _shared;
}

#pragma mark -
#pragma mark Exception Handling

- (void) raiseSqliteException: (NSString *) errorMessage
{
	[NSException raise:@"IDatabaseSQLiteException" format:errorMessage, sqlite3_errmsg16(_database)];
}

#pragma mark -
#pragma mark Open Database

- (BOOL) open
{
	int result = sqlite3_open([self.pathToDatabase cStringUsingEncoding:NSUTF8StringEncoding], &_database);
	
	if(result == SQLITE_OK)
		return YES;
	else 
		return NO;
}

- (NSUInteger) lastInsertRowId
{
	return (NSUInteger) sqlite3_last_insert_rowid(_database);
}

#pragma mark -
#pragma mark Initialization
- (id) init
{
	if(self = [super init])
	{
		[self open];
	}
	
	return self;
}

- (id) initWithPath: (NSString *) filePath
{
	if(self = [super init])
	{
		self.pathToDatabase = filePath;
		[self open];
	}
	
	return self;
}

- (id) initWithFileName: (NSString *) fileName
{
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	
	return [self initWithPath: [documentsDirectory stringByAppendingPathComponent:fileName]];
}

#pragma mark -
#pragma mark SQL
- (int) columnTypeToInt: (NSString *) columnType
{
	if([columnType isEqualToString:@"INTEGER"])
		return SQLITE_INTEGER;
	else if([columnType isEqualToString:@"REAL"])
		return SQLITE_FLOAT;
	else if([columnType isEqualToString:@"TEXT"])
		return SQLITE_TEXT;
	else if ([columnType isEqualToString:@"BLOB"])
		return SQLITE_BLOB;
	else if ([columnType isEqualToString:@"NULL"])
		return SQLITE_NULL;
	
	
	return SQLITE_TEXT;
}

- (int) typeForStatement: (sqlite3_stmt *) statement column: (int) column
{
	const char * columnType = sqlite3_column_decltype(statement, column);
	
	if(columnType != NULL)
	{
		return [self columnTypeToInt:[[NSString stringWithUTF8String: columnType] uppercaseString]];
	}
	
	return sqlite3_column_type(statement, column);
}

- (NSArray *) columnTypesForStatement: (sqlite3_stmt *) statement
{
	int columnCount = sqlite3_column_count(statement);
	
	NSMutableArray *columnTypes = [NSMutableArray array];
	for(int i = 0; i < columnCount; i++)
	{
		[columnTypes addObject:[NSNumber numberWithInt:[self typeForStatement:statement column:i]]];
	}
	
	return columnTypes;
}

- (NSArray *) columnNamesForStatement: (sqlite3_stmt *) statement
{
	int columnCount = sqlite3_column_count(statement);
	
	NSMutableArray *columnNames = [NSMutableArray array];
	for(int i = 0; i < columnCount; i++)
	{
		[columnNames addObject:[NSString stringWithUTF8String:sqlite3_column_name(statement, i)]];
	}
	
	return columnNames;
}

- (void) copyValuesFromStatement: (sqlite3_stmt *) statement toRow: (id) row queryInfo: (NSDictionary *) queryInfo columnTypes: (NSArray *) columnTypes columnNames: (NSArray *) columnNames
{
	int columnCount = sqlite3_column_count(statement);
	
	for(int i = 0; i < columnCount; i++)
	{
		id value = [self valueFromStatement:statement column:i queryInfo: queryInfo columnTypes: columnTypes];
		
		if(value != nil)
		{
			[row setValue: value forKey: [columnNames objectAtIndex:i]];
		}
	}
}

- (id) valueFromStatement: (sqlite3_stmt *) statement column: (int) column queryInfo: (NSDictionary *) queryInfo columnTypes: (NSArray *) columnTypes
{	
	int columnType = [[columnTypes objectAtIndex:column] intValue];
	
	if(columnType == SQLITE_INTEGER)
	{
		return [NSNumber numberWithInt:sqlite3_column_int(statement, column)];
	}
	else if(columnType == SQLITE_FLOAT)
	{
		return [NSNumber numberWithDouble: sqlite3_column_double(statement, column)];
	}
	else if(columnType == SQLITE_TEXT)
	{
		const char *text = (const char *) sqlite3_column_text(statement, column);
		if(text != nil)
			return [NSString stringWithUTF8String: text];
		else
			return nil;
			
	}
	else if (columnType == SQLITE_BLOB)
	{
		return [NSData dataWithBytes:sqlite3_column_blob(statement, column) length:sqlite3_column_bytes(statement, column)];
	}
	else if (columnType == SQLITE_NULL)
	{
		return nil;
	}
	
	NSLog(@"Unrecognized SQL column type: %i for sql: %@", columnType, [queryInfo objectForKey:@"sql"]);
	
	return nil;
}

- (void) bindArguments: (NSArray *) arguments toStatement: (sqlite3_stmt *) statement queryInfo: (NSDictionary *) queryInfo
{
	int expectedArguments = sqlite3_bind_parameter_count(statement);
	
	NSAssert2(expectedArguments == [arguments count], @"Number of bound parameters does not match for sql: %@ parameters: '%@'", 
			  [queryInfo objectForKey:@"sql"], [queryInfo objectForKey:@"parameters"]);
	
	for(int i = 1; i <= expectedArguments; i++)
	{
		id argument = [arguments objectAtIndex:i - 1];
		if([argument isKindOfClass:[NSString class]])
			sqlite3_bind_text(statement, i, [argument UTF8String], -1, SQLITE_TRANSIENT);
		else if([argument isKindOfClass:[NSData class]])
			sqlite3_bind_blob(statement, i, [argument bytes], [argument length], SQLITE_TRANSIENT);
		else if([argument isKindOfClass:[NSDate class]])
			sqlite3_bind_double(statement, i, [argument timeIntervalSince1970]);
		else if([argument isKindOfClass:[NSNumber class]])
			sqlite3_bind_double(statement, i, [argument doubleValue]);
		else if([argument isKindOfClass:[NSNull class]])
			sqlite3_bind_null(statement, i);
		else
		{
			sqlite3_finalize(statement);
			[NSException raise:@"Unrecognized object type" format:@"Active record doesn't know how to handle object:'%@' bound to sql: %@ position: %i", 
			 argument, [queryInfo objectForKey:@"sql"], i];
		}
	}
}

- (NSArray *) executeSql: (NSString *) sql withParameters: (NSArray *) parameters withClassForRow: (Class) rowClass
{
	NSMutableDictionary *queryInfo = [NSMutableDictionary dictionary];
	[queryInfo setObject:sql forKey:@"sql"];
	
	if(parameters == nil)
		parameters = [NSArray array];
	
	[queryInfo setObject:parameters forKey:@"parameters"];
	
	NSMutableArray *rows = [NSMutableArray array];
	
	sqlite3_stmt *statement;
	const char *iSql = [sql cStringUsingEncoding:NSUTF8StringEncoding];
	
	if(sqlite3_prepare_v2(_database, iSql, -1, &statement, NULL) == SQLITE_OK)
	{
		[self bindArguments:parameters toStatement:statement queryInfo:queryInfo];
		
		BOOL needsToFetchColumnTypesAndNames = YES;
		NSArray *columnTypes = nil;
		NSArray *columnNames = nil;
		
		//if(sqlite3_exec(_database, iSql, nil, &statement, NULL) == SQLITE_OK)
		//{
			while (sqlite3_step(statement) == SQLITE_ROW)
			{
				if(needsToFetchColumnTypesAndNames)
				{
					columnTypes = [self columnTypesForStatement:statement];
					columnNames = [self columnNamesForStatement:statement];
					needsToFetchColumnTypesAndNames = NO;
				}
				
				id row = [[rowClass alloc] init];
				[self copyValuesFromStatement:statement toRow:row queryInfo:queryInfo columnTypes:columnTypes columnNames:columnNames];
				[rows addObject:row];
			}
		//}
	}
	else
	{
		NSLog(@"SQL Warning: failed to prepare statement with message '%s'.", sqlite3_errmsg(_database));
		sqlite3_finalize(statement);
		[self raiseSqliteException: [[NSString stringWithFormat:@"failed to execute statement: '%@', parameters: '%@' with message: ", sql, parameters] stringByAppendingString:@"%S"]];
	}
	
	sqlite3_finalize(statement);
	return rows;
}

- (NSArray *) executeSqlWithParameters: (NSString *) sql, ...
{
	NSMutableArray *arguments = [NSMutableArray array];

	va_list args;
    va_start(args, sql);
	
	id arg;
	while(arg = va_arg(args, id))
		[arguments addObject:arg];

	va_end(args);	
	
	return [self executeSql:sql withParameters: arguments];
}

- (NSArray *) executeSql: (NSString *) sql withParameters: (NSArray *) parameters
{
	return [self executeSql:sql withParameters:parameters withClassForRow: [NSMutableDictionary class]];
}

- (NSArray *) executeSql: (NSString *) sql
{
	return [self executeSql:sql withParameters: nil];
}

#pragma mark -
#pragma mark Table Info
- (NSArray *) tables
{
	return [self executeSql:@"select * from sqlite_master where type = 'table'"];
}

- (NSArray *) tableNames
{
	return [[self tables] valueForKey:@"name"];
}

- (NSArray *) columnsForTableName: (NSString *) tableName
{	
	NSArray *results = [self executeSql: [NSString stringWithFormat:@"pragma table_info(%@)", tableName]];
	
	return [results valueForKey:@"name"];
}

@end
